;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "verb" "20260414.1812"
  "Organize and send HTTP requests."
  '((emacs "26.3"))
  :url "https://github.com/federicotdn/verb"
  :commit "ffe668c20120ba215f14d6d5720e327fbadbd10d"
  :revdesc "ffe668c20120"
  :keywords '("tools")
  :authors '(("Federico Tedin" . "federicotedin@gmail.com"))
  :maintainers '(("Federico Tedin" . "federicotedin@gmail.com")))
